@@ -1,4 +1,5 @@
 use crate::config::Config;
+use crate::config::Layout;
 use crate::config::Mode;
 use crate::input::Key;
 use anyhow::{bail, Result};
@@ -11,7 +12,6 @@ use std::collections::VecDeque;
 use std::env;
 use std::fs;
 use std::io;
-use std::io::prelude::*;
 use std::path::PathBuf;
 
 pub const TEMPLATE_TABLE_ROW: &str = "TEMPLATE_TABLE_ROW";
@@ -24,7 +24,6 @@ pub struct Pipe {
     focus_out: String,
     selection_out: String,
     result_out: String,
-    mode_out: String,
     directory_nodes_out: String,
     global_help_menu_out: String,
     logs_out: String,
@@ -45,8 +44,6 @@ impl Pipe {
 
         let result_out = pipesdir.join("result_out").to_string_lossy().to_string();
 
-        let mode_out = pipesdir.join("mode_out").to_string_lossy().to_string();
-
         let directory_nodes_out = pipesdir
             .join("directory_nodes_out")
             .to_string_lossy()
@@ -64,7 +61,6 @@ impl Pipe {
         fs::write(&msg_in, "")?;
         fs::write(&focus_out, "")?;
         fs::write(&selection_out, "")?;
-        fs::write(&mode_out, "")?;
         fs::write(&directory_nodes_out, "")?;
         fs::write(&global_help_menu_out, "")?;
         fs::write(&result_out, "")?;
@@ -76,7 +72,6 @@ impl Pipe {
             focus_out,
             selection_out,
             result_out,
-            mode_out,
             directory_nodes_out,
             global_help_menu_out,
             logs_out,
@@ -104,11 +99,6 @@ impl Pipe {
         &self.result_out
     }
 
-    /// Get a reference to the pipe's mode out.
-    pub fn mode_out(&self) -> &String {
-        &self.mode_out
-    }
-
     /// Get a reference to the pipe's directory nodes out.
     pub fn directory_nodes_out(&self) -> &String {
         &self.directory_nodes_out
@@ -1078,6 +1068,27 @@ pub enum ExternalMsg {
     /// **Example:** `SwitchModeCustom: my_custom_mode`
     SwitchModeCustom(String),
 
+    /// Switch layout.
+    /// This will call `Refresh` automatically.
+    ///
+    /// > **NOTE:** To be specific about which layout to switch to, use `SwitchLayoutBuiltin` or
+    /// `SwitchLayoutCustom` instead.
+    ///
+    /// **Example:** `SwitchLayout: default`
+    SwitchLayout(String),
+
+    /// Switch to a builtin layout.
+    /// This will call `Refresh` automatically.
+    ///
+    /// **Example:** `SwitchLayoutBuiltin: default`
+    SwitchLayoutBuiltin(String),
+
+    /// Switch to a custom layout.
+    /// This will call `Refresh` automatically.
+    ///
+    /// **Example:** `SwitchLayoutCustom: my_custom_layout`
+    SwitchLayoutCustom(String),
+
     /// Call a shell command with the given arguments.
     /// Note that the arguments will be shell-escaped.
     /// So to read the variables, the `-c` option of the shell
@@ -1396,6 +1407,7 @@ pub struct App {
     selection: IndexSet<Node>,
     msg_out: VecDeque<MsgOut>,
     mode: Mode,
+    layout: Layout,
     input_buffer: Option<String>,
     pid: u32,
     session_path: String,
@@ -1436,7 +1448,13 @@ impl App {
             )
         };
 
-        let mode = match config.modes().builtin().get(&"default".to_string()) {
+        let mode = match config.modes().get(
+            &config
+                .general()
+                .initial_mode()
+                .to_owned()
+                .unwrap_or_else(|| "default".into()),
+        ) {
             Some(m) => m
                 .clone()
                 .sanitized(config.general().read_only().unwrap_or_default()),
@@ -1445,6 +1463,19 @@ impl App {
             }
         };
 
+        let layout = match config.layouts().get(
+            &config
+                .general()
+                .initial_layout()
+                .to_owned()
+                .unwrap_or_else(|| "default".into()),
+        ) {
+            Some(l) => l.clone(),
+            None => {
+                bail!("'default' layout is missing")
+            }
+        };
+
         let pid = std::process::id();
         let session_path = dirs::runtime_dir()
             .unwrap_or_else(|| "/tmp".into())
@@ -1476,6 +1507,7 @@ impl App {
             selection: Default::default(),
             msg_out: Default::default(),
             mode,
+            layout,
             input_buffer: Default::default(),
             pid,
             session_path: session_path.clone(),
@@ -1499,8 +1531,8 @@ impl App {
             ));
         }
 
-        fs::write(&app.pipe().global_help_menu_out, app.global_help_menu_str())?;
-        app.write_pipes(None)
+        app.write_pipes()?;
+        Ok(app)
     }
 
     pub fn focused_node(&self) -> Option<&Node> {
@@ -1518,10 +1550,10 @@ impl App {
         self
     }
 
-    pub fn handle_task(self, task: Task, last_app: &Self) -> Result<Self> {
+    pub fn handle_task(self, task: Task) -> Result<Self> {
         match task.msg {
             MsgIn::Internal(msg) => self.handle_internal(msg),
-            MsgIn::External(msg) => self.handle_external(msg, task.key, last_app),
+            MsgIn::External(msg) => self.handle_external(msg, task.key),
         }
     }
 
@@ -1532,7 +1564,7 @@ impl App {
         }
     }
 
-    fn handle_external(self, msg: ExternalMsg, key: Option<Key>, last_app: &Self) -> Result<Self> {
+    fn handle_external(self, msg: ExternalMsg, key: Option<Key>) -> Result<Self> {
         if self.config().general().read_only().unwrap_or_default() && !msg.is_read_only() {
             self.log_error("Cannot call shell command in read-only mode.".into())
         } else {
@@ -1577,6 +1609,9 @@ impl App {
                 ExternalMsg::SwitchMode(mode) => self.switch_mode(&mode),
                 ExternalMsg::SwitchModeBuiltin(mode) => self.switch_mode_builtin(&mode),
                 ExternalMsg::SwitchModeCustom(mode) => self.switch_mode_custom(&mode),
+                ExternalMsg::SwitchLayout(mode) => self.switch_layout(&mode),
+                ExternalMsg::SwitchLayoutBuiltin(mode) => self.switch_layout_builtin(&mode),
+                ExternalMsg::SwitchLayoutCustom(mode) => self.switch_layout_custom(&mode),
                 ExternalMsg::Call(cmd) => self.call(cmd),
                 ExternalMsg::CallSilently(cmd) => self.call_silently(cmd),
                 ExternalMsg::BashExec(cmd) => self.bash_exec(cmd),
@@ -1618,7 +1653,6 @@ impl App {
             }
         }?
         .refresh_selection()
-        .write_pipes(Some(&last_app))
     }
 
     fn handle_key(mut self, key: Key) -> Result<Self> {
@@ -1949,6 +1983,36 @@ impl App {
         }
     }
 
+    fn switch_layout(mut self, layout: &str) -> Result<Self> {
+        if let Some(l) = self.config().layouts().get(layout) {
+            self.layout = l.to_owned();
+            self.msg_out.push_back(MsgOut::Refresh);
+            Ok(self)
+        } else {
+            self.log_error(format!("Layout not found: {}", layout))
+        }
+    }
+
+    fn switch_layout_builtin(mut self, layout: &str) -> Result<Self> {
+        if let Some(l) = self.config().layouts().get_builtin(layout) {
+            self.layout = l.to_owned();
+            self.msg_out.push_back(MsgOut::Refresh);
+            Ok(self)
+        } else {
+            self.log_error(format!("Builtin layout not found: {}", layout))
+        }
+    }
+
+    fn switch_layout_custom(mut self, layout: &str) -> Result<Self> {
+        if let Some(l) = self.config().layouts().get_custom(layout) {
+            self.layout = l.to_owned();
+            self.msg_out.push_back(MsgOut::Refresh);
+            Ok(self)
+        } else {
+            self.log_error(format!("Custom layout not found: {}", layout))
+        }
+    }
+
     fn call(mut self, command: Command) -> Result<Self> {
         self.msg_out.push_back(MsgOut::Call(command));
         Ok(self)
@@ -2283,10 +2347,10 @@ impl App {
         &self.session_path
     }
 
-    fn refresh_selection(mut self) -> Self {
+    fn refresh_selection(mut self) -> Result<Self> {
         self.selection
             .retain(|n| PathBuf::from(&n.absolute_path).exists());
-        self
+        Ok(self)
     }
 
     pub fn result(&self) -> Vec<&Node> {
@@ -2335,6 +2399,14 @@ impl App {
         &self.logs
     }
 
+    pub fn logs_str(&self) -> String {
+        self.logs()
+            .iter()
+            .map(|l| format!("{}\n", l))
+            .collect::<Vec<String>>()
+            .join("")
+    }
+
     pub fn global_help_menu_str(&self) -> String {
         let builtin = self.config().modes().builtin().clone();
         let custom = self.config().modes().custom().clone();
@@ -2411,7 +2483,7 @@ impl App {
             .join("")
     }
 
-    fn write_pipes(self, last_app: Option<&Self>) -> Result<Self> {
+    pub fn write_pipes(&self) -> Result<()> {
         // TODO optimize and test
 
         let focused_node_str = self.focused_node_str();
@@ -2438,11 +2510,6 @@ impl App {
         fs::write(&self.pipe().history_out, history_str)?;
         // };
 
-        let mode_str = self.mode_str();
-        // if last_app.map(|a| a.mode_str() != mode_str).unwrap_or(true) {
-        fs::write(&self.pipe().mode_out, mode_str)?;
-        // };
-
         let directory_nodes_str = self.directory_nodes_str();
         // if last_app
         //     .map(|a| a.directory_nodes_str() != directory_nodes_str)
@@ -2455,20 +2522,21 @@ impl App {
         //     .map(|a| a.logs().len() != self.logs().len())
         //     .unwrap_or(true)
         // {
-        let new_logs = self
-            .logs()
-            .iter()
-            .skip(last_app.map(|a| a.logs().len()).unwrap_or(0))
-            .map(|l| format!("{}\n", l))
-            .collect::<Vec<String>>()
-            .join("");
+        // let new_logs = self
+        //     .logs()
+        //     .iter()
+        //     .skip(last_app.map(|a| a.logs().len()).unwrap_or(0))
+        //     .map(|l| format!("{}\n", l))
+        //     .collect::<Vec<String>>()
+        //     .join("");
 
-        let mut file = fs::OpenOptions::new()
-            .append(true)
-            .open(&self.pipe().logs_out)?;
-
-        file.write_all(new_logs.as_bytes())?;
+        // let mut file = fs::OpenOptions::new()
+        //     .append(true)
+        //     .open(&self.pipe().logs_out)?;
+        // file.write_all(new_logs.as_bytes())?;
         // };
+        let logs_str = self.logs_str();
+        fs::write(&self.pipe().logs_out, logs_str)?;
 
         let result_str = self.result_str();
         // if last_app
@@ -2477,6 +2545,11 @@ impl App {
         // {
         fs::write(&self.pipe().result_out, result_str)?;
         // };
-        Ok(self)
+        Ok(())
+    }
+
+    /// Get a reference to the app's layout.
+    pub fn layout(&self) -> &Layout {
+        &self.layout
     }
 }
