@@ -4,6 +4,7 @@ use crate::app::NodeFilter;
 use crate::app::NodeSorter;
 use crate::app::NodeSorterApplicable;
 use crate::default_config;
+use crate::ui::Border;
 use crate::ui::Style;
 use anyhow::Result;
 use indexmap::IndexSet;
@@ -11,6 +12,7 @@ use serde::{Deserialize, Serialize};
 use std::collections::BTreeMap;
 use std::collections::HashMap;
 use tui::layout::Constraint as TuiConstraint;
+use tui::layout::Rect;
 
 #[derive(Debug, Clone, PartialEq, Default, Serialize, Deserialize)]
 #[serde(deny_unknown_fields)]
@@ -72,8 +74,8 @@ impl NodeTypeConfig {
     }
 
     /// Get a reference to the node type config's style.
-    pub fn style(&self) -> Style {
-        self.style
+    pub fn style(&self) -> &Style {
+        &self.style
     }
 
     /// Get a reference to the node type config's meta.
@@ -178,8 +180,8 @@ impl UiConfig {
     }
 
     /// Get a reference to the ui config's style.
-    pub fn style(&self) -> Style {
-        self.style
+    pub fn style(&self) -> &Style {
+        &self.style
     }
 }
 
@@ -194,7 +196,7 @@ pub struct UiElement {
 }
 
 impl UiElement {
-    fn extend(mut self, other: Self) -> Self {
+    pub fn extend(mut self, other: Self) -> Self {
         self.format = other.format.or(self.format);
         self.style = self.style.extend(other.style);
         self
@@ -206,8 +208,8 @@ impl UiElement {
     }
 
     /// Get a reference to the ui element's style.
-    pub fn style(&self) -> Style {
-        self.style
+    pub fn style(&self) -> &Style {
+        &self.style
     }
 }
 
@@ -238,8 +240,8 @@ impl TableRowConfig {
     }
 
     /// Get a reference to the table row config's style.
-    pub fn style(&self) -> Style {
-        self.style
+    pub fn style(&self) -> &Style {
+        &self.style
     }
 
     /// Get a reference to the table row config's height.
@@ -249,29 +251,55 @@ impl TableRowConfig {
 }
 
 #[derive(Debug, Clone, Copy, Serialize, Deserialize)]
-#[serde(rename_all = "snake_case", deny_unknown_fields)]
+#[serde(deny_unknown_fields)]
 pub enum Constraint {
     Percentage(u16),
     Ratio(u32, u32),
     Length(u16),
+    LengthLessThanScreenHeight(u16),
+    LengthLessThanScreenWidth(u16),
+    LengthLessThanLayoutHeight(u16),
+    LengthLessThanLayoutWidth(u16),
     Max(u16),
+    MaxLessThanScreenHeight(u16),
+    MaxLessThanScreenWidth(u16),
+    MaxLessThanLayoutHeight(u16),
+    MaxthLessThanLayoutWidth(u16),
     Min(u16),
+    MinLessThanScreenHeight(u16),
+    MinLessThanScreenWidth(u16),
+    MinLessThanLayoutHeight(u16),
+    MinLessThanLayoutWidth(u16),
 }
 
-impl Default for Constraint {
-    fn default() -> Self {
-        Self::Min(1)
-    }
-}
-
-impl Into<TuiConstraint> for Constraint {
-    fn into(self) -> TuiConstraint {
+impl Constraint {
+    pub fn to_tui(self, screen_size: Rect, layout_size: Rect) -> TuiConstraint {
         match self {
-            Self::Length(n) => TuiConstraint::Length(n),
             Self::Percentage(n) => TuiConstraint::Percentage(n),
             Self::Ratio(x, y) => TuiConstraint::Ratio(x, y),
+            Self::Length(n) => TuiConstraint::Length(n),
+            Self::LengthLessThanScreenHeight(n) => {
+                TuiConstraint::Length(screen_size.height.max(n) - n)
+            }
+            Self::LengthLessThanScreenWidth(n) => {
+                TuiConstraint::Length(screen_size.width.max(n) - n)
+            }
+            Self::LengthLessThanLayoutHeight(n) => {
+                TuiConstraint::Length(layout_size.height.max(n) - n)
+            }
+            Self::LengthLessThanLayoutWidth(n) => {
+                TuiConstraint::Length(layout_size.width.max(n) - n)
+            }
             Self::Max(n) => TuiConstraint::Max(n),
+            Self::MaxLessThanScreenHeight(n) => TuiConstraint::Max(screen_size.height.max(n) - n),
+            Self::MaxLessThanScreenWidth(n) => TuiConstraint::Max(screen_size.width.max(n) - n),
+            Self::MaxLessThanLayoutHeight(n) => TuiConstraint::Max(layout_size.height.max(n) - n),
+            Self::MaxthLessThanLayoutWidth(n) => TuiConstraint::Max(layout_size.width.max(n) - n),
             Self::Min(n) => TuiConstraint::Min(n),
+            Self::MinLessThanScreenHeight(n) => TuiConstraint::Min(screen_size.height.max(n) - n),
+            Self::MinLessThanScreenWidth(n) => TuiConstraint::Min(screen_size.width.max(n) - n),
+            Self::MinLessThanLayoutHeight(n) => TuiConstraint::Min(layout_size.height.max(n) - n),
+            Self::MinLessThanLayoutWidth(n) => TuiConstraint::Min(layout_size.width.max(n) - n),
         }
     }
 }
@@ -320,8 +348,8 @@ impl TableConfig {
     }
 
     /// Get a reference to the table config's style.
-    pub fn style(&self) -> Style {
-        self.style
+    pub fn style(&self) -> &Style {
+        &self.style
     }
 
     /// Get a reference to the table config's tree.
@@ -411,6 +439,9 @@ pub struct SortAndFilterUi {
     #[serde(default)]
     separator: UiElement,
 
+    #[serde(default)]
+    default_identifier: UiElement,
+
     #[serde(default)]
     sort_direction_identifiers: SortDirectionIdentifiersUi,
 
@@ -424,6 +455,7 @@ pub struct SortAndFilterUi {
 impl SortAndFilterUi {
     pub fn extend(mut self, other: Self) -> Self {
         self.separator = self.separator.extend(other.separator);
+        self.default_identifier = self.default_identifier.extend(other.default_identifier);
         self.sort_direction_identifiers = self
             .sort_direction_identifiers
             .extend(other.sort_direction_identifiers);
@@ -451,6 +483,75 @@ impl SortAndFilterUi {
     pub fn filter_identifiers(&self) -> &HashMap<NodeFilter, UiElement> {
         &self.filter_identifiers
     }
+
+    /// Get a reference to the sort and filter ui's default identifier.
+    pub fn default_identifier(&self) -> &UiElement {
+        &self.default_identifier
+    }
+}
+
+#[derive(Debug, Clone, Default, Serialize, Deserialize)]
+#[serde(deny_unknown_fields)]
+pub struct PanelUi {
+    #[serde(default)]
+    default: PanelUiConfig,
+
+    #[serde(default)]
+    table: PanelUiConfig,
+
+    #[serde(default)]
+    sort_and_filter: PanelUiConfig,
+
+    #[serde(default)]
+    selection: PanelUiConfig,
+
+    #[serde(default)]
+    input_and_logs: PanelUiConfig,
+
+    #[serde(default)]
+    help_menu: PanelUiConfig,
+}
+
+impl PanelUi {
+    fn extend(mut self, other: Self) -> Self {
+        self.default = self.default.extend(other.default);
+        self.table = self.table.extend(other.table);
+        self.sort_and_filter = self.sort_and_filter.extend(other.sort_and_filter);
+        self.selection = self.selection.extend(other.selection);
+        self.input_and_logs = self.input_and_logs.extend(other.input_and_logs);
+        self.help_menu = self.help_menu.extend(other.help_menu);
+        self
+    }
+
+    /// Get a reference to the panel ui's default.
+    pub fn default(&self) -> &PanelUiConfig {
+        &self.default
+    }
+
+    /// Get a reference to the panel ui's table.
+    pub fn table(&self) -> &PanelUiConfig {
+        &self.table
+    }
+
+    /// Get a reference to the panel ui's sort and filter.
+    pub fn sort_and_filter(&self) -> &PanelUiConfig {
+        &self.sort_and_filter
+    }
+
+    /// Get a reference to the panel ui's selection.
+    pub fn selection(&self) -> &PanelUiConfig {
+        &self.selection
+    }
+
+    /// Get a reference to the panel ui's input and log.
+    pub fn input_and_logs(&self) -> &PanelUiConfig {
+        &self.input_and_logs
+    }
+
+    /// Get a reference to the panel ui's help menu.
+    pub fn help_menu(&self) -> &PanelUiConfig {
+        &self.help_menu
+    }
 }
 
 #[derive(Debug, Clone, Default, Serialize, Deserialize)]
@@ -486,8 +587,17 @@ pub struct GeneralConfig {
     #[serde(default)]
     sort_and_filter_ui: SortAndFilterUi,
 
+    #[serde(default)]
+    panel_ui: PanelUi,
+
     #[serde(default)]
     initial_sorting: Option<IndexSet<NodeSorterApplicable>>,
+
+    #[serde(default)]
+    initial_mode: Option<String>,
+
+    #[serde(default)]
+    initial_layout: Option<String>,
 }
 
 impl GeneralConfig {
@@ -502,7 +612,10 @@ impl GeneralConfig {
         self.focus_ui = self.focus_ui.extend(other.focus_ui);
         self.selection_ui = self.selection_ui.extend(other.selection_ui);
         self.sort_and_filter_ui = self.sort_and_filter_ui.extend(other.sort_and_filter_ui);
+        self.panel_ui = self.panel_ui.extend(other.panel_ui);
         self.initial_sorting = other.initial_sorting.or(self.initial_sorting);
+        self.initial_layout = other.initial_layout.or(self.initial_layout);
+        self.initial_mode = other.initial_mode.or(self.initial_mode);
         self
     }
 
@@ -560,6 +673,21 @@ impl GeneralConfig {
     pub fn initial_sorting(&self) -> &Option<IndexSet<NodeSorterApplicable>> {
         &self.initial_sorting
     }
+
+    /// Get a reference to the general config's initial mode.
+    pub fn initial_mode(&self) -> &Option<String> {
+        &self.initial_mode
+    }
+
+    /// Get a reference to the general config's initial layout.
+    pub fn initial_layout(&self) -> &Option<String> {
+        &self.initial_layout
+    }
+
+    /// Get a reference to the general config's panel ui.
+    pub fn panel_ui(&self) -> &PanelUi {
+        &self.panel_ui
+    }
 }
 
 #[derive(Debug, Clone, Default, Serialize, Deserialize)]
@@ -605,10 +733,8 @@ impl KeyBindings {
                 .into_iter()
                 .filter(|(_, v)| self.on_key.contains_key(v))
                 .collect();
-            self
-        } else {
-            self
-        }
+        };
+        self
     }
 
     fn extend(mut self, other: Self) -> Self {
@@ -808,6 +934,9 @@ pub struct BuiltinModesConfig {
 
     #[serde(default)]
     sort: Mode,
+
+    #[serde(default)]
+    switch_layout: Mode,
 }
 
 impl BuiltinModesConfig {
@@ -831,6 +960,7 @@ impl BuiltinModesConfig {
             .relative_path_does_not_contain
             .extend(other.relative_path_does_not_contain);
         self.sort = self.sort.extend(other.sort);
+        self.switch_layout = self.switch_layout.extend(other.switch_layout);
         self
     }
 
@@ -857,6 +987,8 @@ impl BuiltinModesConfig {
             "relative path does contain" => Some(&self.relative_path_does_contain),
             "relative_path_does_not_contain" => Some(&self.relative_path_does_not_contain),
             "relative path does not contain" => Some(&self.relative_path_does_not_contain),
+            "switch layout" => Some(&self.switch_layout),
+            "switch_layout" => Some(&self.switch_layout),
             _ => None,
         }
     }
@@ -977,11 +1109,243 @@ impl ModesConfig {
     }
 }
 
+#[derive(Debug, Clone, Default, Serialize, Deserialize)]
+#[serde(deny_unknown_fields)]
+pub struct LayoutOptions {
+    #[serde(default)]
+    margin: Option<u16>,
+    #[serde(default)]
+    horizontal_margin: Option<u16>,
+    #[serde(default)]
+    vertical_margin: Option<u16>,
+    #[serde(default)]
+    constraints: Option<Vec<Constraint>>,
+}
+
+impl LayoutOptions {
+    pub fn extend(mut self, other: Self) -> Self {
+        self.margin = other.margin.or(self.margin);
+        self.horizontal_margin = other.horizontal_margin.or(self.horizontal_margin);
+        self.vertical_margin = other.vertical_margin.or(self.vertical_margin);
+        self.constraints = other.constraints.or(self.constraints);
+        self
+    }
+
+    /// Get a reference to the layout options's constraints.
+    pub fn constraints(&self) -> &Option<Vec<Constraint>> {
+        &self.constraints
+    }
+
+    /// Get a reference to the layout options's margin.
+    pub fn margin(&self) -> Option<u16> {
+        self.margin
+    }
+
+    /// Get a reference to the layout options's horizontal margin.
+    pub fn horizontal_margin(&self) -> Option<u16> {
+        self.horizontal_margin
+    }
+
+    /// Get a reference to the layout options's vertical margin.
+    pub fn vertical_margin(&self) -> Option<u16> {
+        self.vertical_margin
+    }
+}
+
+#[derive(Debug, Clone, Default, Serialize, Deserialize)]
+#[serde(deny_unknown_fields)]
+pub struct PanelUiConfig {
+    #[serde(default)]
+    title: UiElement,
+
+    #[serde(default)]
+    borders: Option<IndexSet<Border>>,
+
+    #[serde(default)]
+    style: Style,
+}
+
+impl PanelUiConfig {
+    pub fn extend(mut self, other: Self) -> Self {
+        self.title = self.title.extend(other.title);
+        self.borders = other.borders.or(self.borders);
+        self.style = self.style.extend(other.style);
+        self
+    }
+    /// Get a reference to the block config's borders.
+    pub fn borders(&self) -> &Option<IndexSet<Border>> {
+        &self.borders
+    }
+
+    /// Get a reference to the block config's title.
+    pub fn title(&self) -> &UiElement {
+        &self.title
+    }
+
+    /// Get a reference to the block config's style.
+    pub fn style(&self) -> &Style {
+        &self.style
+    }
+}
+
+#[derive(Debug, Clone, Serialize, Deserialize)]
+#[serde(deny_unknown_fields)]
+pub enum Layout {
+    Nothing(Option<PanelUiConfig>),
+    Table(Option<PanelUiConfig>),
+    InputAndLogs(Option<PanelUiConfig>),
+    Selection(Option<PanelUiConfig>),
+    HelpMenu(Option<PanelUiConfig>),
+    SortAndFilter(Option<PanelUiConfig>),
+    Horizontal {
+        config: LayoutOptions,
+        splits: Vec<Layout>,
+    },
+    Vertical {
+        config: LayoutOptions,
+        splits: Vec<Layout>,
+    },
+}
+
+impl Default for Layout {
+    fn default() -> Self {
+        Self::Nothing(Default::default())
+    }
+}
+
+impl Layout {
+    pub fn extend(self, other: Self) -> Self {
+        match (self, other) {
+            (s, Self::Nothing(_)) => s,
+            (Self::Table(s), Self::Table(o)) => Self::Table(o.or(s)),
+            (Self::InputAndLogs(s), Self::InputAndLogs(o)) => Self::InputAndLogs(o.or(s)),
+            (Self::Selection(s), Self::Selection(o)) => Self::Selection(o.or(s)),
+            (Self::HelpMenu(s), Self::HelpMenu(o)) => Self::HelpMenu(o.or(s)),
+            (Self::SortAndFilter(s), Self::SortAndFilter(o)) => Self::SortAndFilter(o.or(s)),
+            (
+                Self::Horizontal {
+                    config: sconfig,
+                    splits: _,
+                },
+                Self::Horizontal {
+                    config: oconfig,
+                    splits: osplits,
+                },
+            ) => Self::Horizontal {
+                config: sconfig.extend(oconfig),
+                splits: osplits,
+            },
+
+            (
+                Self::Vertical {
+                    config: sconfig,
+                    splits: _,
+                },
+                Self::Vertical {
+                    config: oconfig,
+                    splits: osplits,
+                },
+            ) => Self::Vertical {
+                config: sconfig.extend(oconfig),
+                splits: osplits,
+            },
+            (_, other) => other,
+        }
+    }
+}
+
+#[derive(Debug, Clone, Default, Serialize, Deserialize)]
+#[serde(deny_unknown_fields)]
+pub struct BuiltinLayoutsConfig {
+    #[serde(default)]
+    default: Layout,
+
+    #[serde(default)]
+    no_help: Layout,
+
+    #[serde(default)]
+    no_selection: Layout,
+
+    #[serde(default)]
+    no_help_no_selection: Layout,
+}
+
+impl BuiltinLayoutsConfig {
+    pub fn extend(mut self, other: Self) -> Self {
+        self.default = self.default.extend(other.default);
+        self.no_help = self.no_help.extend(other.no_help);
+        self.no_selection = self.no_selection.extend(other.no_selection);
+        self.no_help_no_selection = self.no_help_no_selection.extend(other.no_help_no_selection);
+        self
+    }
+
+    pub fn get(&self, name: &str) -> Option<&Layout> {
+        match name {
+            "default" => Some(&self.default),
+            "no_help" => Some(&self.no_help),
+            "no help" => Some(&self.no_help),
+            "no_selection" => Some(&self.no_selection),
+            "no selection" => Some(&self.no_selection),
+            "no_help_no_selection" => Some(&self.no_help_no_selection),
+            "no help no selection" => Some(&self.no_help_no_selection),
+            _ => None,
+        }
+    }
+
+    /// Get a reference to the builtin layouts config's default.
+    pub fn default(&self) -> &Layout {
+        &self.default
+    }
+}
+
+#[derive(Debug, Clone, Default, Serialize, Deserialize)]
+#[serde(deny_unknown_fields)]
+pub struct LayoutsConfig {
+    #[serde(default)]
+    builtin: BuiltinLayoutsConfig,
+
+    #[serde(default)]
+    custom: HashMap<String, Layout>,
+}
+
+impl LayoutsConfig {
+    pub fn get_builtin(&self, name: &str) -> Option<&Layout> {
+        self.builtin.get(name)
+    }
+
+    pub fn get_custom(&self, name: &str) -> Option<&Layout> {
+        self.custom.get(name)
+    }
+
+    pub fn get(&self, name: &str) -> Option<&Layout> {
+        self.get_builtin(name).or_else(|| self.get_custom(name))
+    }
+
+    pub fn extend(mut self, other: Self) -> Self {
+        self.builtin = self.builtin.extend(other.builtin);
+        self.custom.extend(other.custom);
+        self
+    }
+
+    /// Get a reference to the layouts config's builtin.
+    pub fn builtin(&self) -> &BuiltinLayoutsConfig {
+        &self.builtin
+    }
+
+    /// Get a reference to the layouts config's custom.
+    pub fn custom(&self) -> &HashMap<String, Layout> {
+        &self.custom
+    }
+}
+
 #[derive(Debug, Clone, Serialize, Deserialize)]
 #[serde(deny_unknown_fields)]
 pub struct Config {
     version: String,
 
+    #[serde(default)]
+    layouts: LayoutsConfig,
+
     #[serde(default)]
     general: GeneralConfig,
 
@@ -996,6 +1360,7 @@ impl Default for Config {
     fn default() -> Self {
         Self {
             version: default_config::version(),
+            layouts: default_config::layouts(),
             general: default_config::general(),
             node_types: default_config::node_types(),
             modes: default_config::modes(),
@@ -1006,6 +1371,7 @@ impl Default for Config {
 impl Config {
     pub fn extended(mut self) -> Self {
         let default = Self::default();
+        self.layouts = default.layouts.extend(self.layouts);
         self.general = default.general.extend(self.general);
         self.node_types = default.node_types.extend(self.node_types);
         self.modes = default.modes.extend(self.modes);
@@ -1028,20 +1394,8 @@ impl Config {
 
     pub fn is_compatible(&self) -> Result<bool> {
         let result = match self.parsed_version()? {
-            (0, 5, 13) => true,
-            (0, 5, 12) => true,
-            (0, 5, 11) => true,
-            (0, 5, 10) => true,
-            (0, 5, 9) => true,
-            (0, 5, 8) => true,
-            (0, 5, 7) => true,
-            (0, 5, 6) => true,
-            (0, 5, 5) => true,
-            (0, 5, 4) => true,
-            (0, 5, 3) => true,
-            (0, 5, 2) => true,
-            (0, 5, 1) => true,
-            (0, 5, 0) => true,
+            (0, 7, 1) => true,
+            (0, 7, 0) => true,
             (_, _, _) => false,
         };
 
@@ -1050,20 +1404,11 @@ impl Config {
 
     pub fn upgrade_notification(&self) -> Result<Option<&str>> {
         let result = match self.parsed_version()? {
-            (0, 5, 13) => None,
-            (0, 5, 12) => Some("App version updated. Added new messages for switching mode."),
-            (0, 5, 11) => Some("App version updated. Fixed changing directory on focus using argument"),
-            (0, 5, 10) => Some("App version updated. Added xplr desktop icon and search navigation"),
-            (0, 5, 9) => Some("App version updated. Fixed pipes not updating properly"),
-            (0, 5, 8) => Some("App version updated. Fixed support for filenames starting with - (hiphen)"),
-            (0, 5, 7) => Some("App version updated. Fixed distorted screen when opening files in GUI"),
-            (0, 5, 6) => Some("App version updated. Fixed piping and in-built terminal support"),
-            (0, 5, 5) => Some("App version updated. Significant reduction in CPU usage"),
-            (0, 5, 4) => Some("App version updated. Significant reduction in CPU usage"),
-            (0, 5, 3) => Some("App version updated. Fixed exit on permission denied"),
-            (0, 5, 2) => Some("App version updated. Now pwd is synced with your terminal session"),
-            (0, 5, 1) => Some("App version updated. Now follow symlinks using 'gf'"),
-            (_, _, _) => Some("App version updated. New: added sort and filter support and some hacks: https://github.com/sayanarijit/xplr/wiki/Hacks"),
+            (0, 7, 1) => None,
+            (0, 7, 0) => {
+                Some("App version updated. Use `tab` to select files while in search mode")
+            }
+            (_, _, _) => None,
         };
 
         Ok(result)
@@ -1074,6 +1419,11 @@ impl Config {
         &self.version
     }
 
+    /// Get a reference to the config's layouts.
+    pub fn layouts(&self) -> &LayoutsConfig {
+        &self.layouts
+    }
+
     /// Get a reference to the config's general.
     pub fn general(&self) -> &GeneralConfig {
         &self.general
